#include <string>
#include <stdexcept>
#include <vector>
#include <map>

/** Relational Database Implementation 
    @see http://en.wikipedia.org/wiki/Relational_database

    Nomenclature: Attribute = Column
                  Tuple = Entry
                  DataBase = Relation
*/
class DataBase {
public:

  /// Basic exception type for the DataBase class
  struct DataBaseException : public std::runtime_error{
    DataBaseException(const std::string &functionName, const std::string &errorDescription):
      std::runtime_error(functionName+": "+errorDescription){}
  };
  
  /// Utility macro for instantiation of exception classes
#define DECLARE_DATABASE_EXCEPTION_TYPE(X)      \
  struct X##Exception : public DataBaseException{                       \
    X##Exception(const std::string &functionName, const std::string &errorDescription): \
    DataBaseException(functionName,errorDescription){}                  \
  }
  
  /// instantiation of exception types
  DECLARE_DATABASE_EXCEPTION_TYPE(InvalidID);
  DECLARE_DATABASE_EXCEPTION_TYPE(InvalidRowName);
  DECLARE_DATABASE_EXCEPTION_TYPE(InvalidType);
  DECLARE_DATABASE_EXCEPTION_TYPE(InvalidName);
  DECLARE_DATABASE_EXCEPTION_TYPE(FileIO);
  DECLARE_DATABASE_EXCEPTION_TYPE(DoubledEntry);


  /// typedef for a single table entry
  typedef std::pair<std::string,std::string> Entry;

  /// typedef for a list of entries
  typedef std::map<std::string,std::string> EntryList;
  
private:
  /// data storage details have to be hidden (encapsulated in this class)
  class Data; // siehe hierzu auch den Kommentar am Ende der Datei

  /// Internal instance for data storage (allocated in constructors, deallocated in destructor)
  Data *m_data;

public:

  /// default constructor
  /** Creates an empty database with unremovable default columns 'surname', 'first name' and 'id' 
      'surname' and 'first name' are of type 'nonEmptyString' and id is of type 'positive int'
      */
  DataBase();

  /// Copy constructor (creates a deep copy of whole data base)
  DataBase(const DataBase &base);

  /// Assignment operator (creates a deep copy of whole data base)
  DataBase &operator=(const DataBase &base);
  
  /// Destructor
  ~DataBase();
  
  /// adds a new entry with given name, returns unique ID for the new entry
  /** @param surname new person's surname 
      @param firsName new person's first name
      @param entries map of key/value pairs (each must be checked if valid)
  */
  unsigned int addEntry(const std::string &surname, const std::string &firstName, const EntryList &entries);

  /// removes all entries that match the given filter
  /** The filter can be applied using the query method 
      returns the count of actually removed rows
  */
  int removeEntry(const EntryList &filter);
  
  /// removes entry with given ID
  /** An exception is thrown if given ID is not valid */
  void removeEntryByID(unsigned int id) throw (InvalidIDException);

  /// removes all current entries for the database
  void removeAllEntries();

  /// adds a new column to the data base
  /** Type may be
      - string
      - nonEmptyString
      - int
      - unsigned int
      - positive int
      - float
      
      Exception is thrown if column name does already exist or if type is not valid
      column-names are case-insensitive
  */
  void addColumn(const std::string &name, const std::string &type) throw (InvalidRowNameException,InvalidTypeException);
  
  /// removes the row with given row-name
  /* The rows 'first name', 'surname' and 'id' must not be removed!. If the user tries to remove one of these
     Columns, an std::logic_error is thrown.
  */
  void removeColumn(const std::string &name) throw (InvalidRowNameException, std::logic_error);


  /// Dumps the whole database content into a file
  /** File format should be human readable and human editable */
  void dumpToFile(const std::string &filename) const throw (FileIOException);

  /// grabs database entries from dump-file
  /** @param filename input database file name 
      @param joinWithCurrentContent if true, old content is not lost (see also join!)
  */
  void readFromDump(const std::string &filename, bool joinWithCurrentContent=false) 
    throw (FileIOException, InvalidRowNameException, std::logic_error, InvalidTypeException, DoubledEntryException);

  /// joins this database with another one
  /** First name, surname combinations must not be doubled.  New entries have to get a new ID */
  void join(const DataBase &other) throw (DoubledEntryException);

  /// returns a list of currently contained columns
  const std::vector<std::string> getColumns() const;
  
  /// complex query-function
  /** Returns all rows that match given filters
      Example:
      'ID'  'surname' 'first name' 'street'       'rating'
       1     Mueller   Klaus        Bahnhofstr.    0.5
       2     Mueller   Fritz        Bahnhofstr.    0.7
       3     Schmidt   Erna         Hauptstr.      0.8
       4     Schulze   Erna         Bahnforstr.    0.8

      Query 1:
      Filters = [ (SURNAME,Mueller) ] // note: column names are case-insensitive
      Result: [ (ID,1) (surname,Mueller) (first name,Klaus) , (street,Bahnhofstr.) (rating,0.5) ]
              [ (ID,2) (surname,Mueller) (first name,Fritz) , (street,Bahnhofstr.) (rating,0.7) ]  

      Note:
      [ ... ] is one EntryList that has key/value entries (key,value)
              keys are column-names values are cell-entries of the corresponding row
      

      Query 2:
      Filters = [ (rating,0.8) (surname,Schulze) ] 
      Result: [ (ID,4) (surname,Schulze) (first name,Erna) , (street,Bahnhofstr.) (rating,0.8) ]

      An Empty 'filters'-list will match nothing!
  */
  const std::vector<EntryList> query(const EntryList &filters) const;

  /// shortcut for query by ID (as ID is unique, only a single row is returned)
  EntryList queryByID(unsigned int ID) const throw (InvalidIDException);

  /// shortcut for query by given full name
  /** as name combinations are unique, only a single row is returned */
  EntryList queryByName(const std::string &surname, const std::string &firstName) const throw (InvalidNameException);
  
  /// returns all entries of this database
  const std::vector<EntryList> queryAll() const;
};


/**********************************************************************************************************

Anmerkung zum Thema private implementation (pimpl)

Für komplexe Klassen wie die DataBase-Klasse, wird oft das sog. pimpl-pattern verwendet. Die Idee hierbei
ist, dass nicht nur die Implementation der Klasse ausschließlich in der C++-Datei verfügbar ist, sondern,
dass zusätzlich auch das Daten-Management der Klasse in eine C++-Datei ausgelagert wird. Aus diesem Grund hat
die eigentliche Klasse nur eine Klassen-Vorausdeklaration (hier 'class Data') und einen Pointer auf die
Daten (Data* m_data). In der C++-Datei wird dann zunächst die Definition der Klasse nachgeliefert:

// hier kann es auch ein struct sein->zugriff ist ja eh nur lokal
struct DataBase::Data{
   // columns  }
   // rows     } -- oder ähnlich
   // content  }
};

Dann kann die Data klasse in der DataBase Implementation verwendet werden:

DataBase::DataBase():m_data(new DataBase::Data){
    // m_data->colums.add ... -- oder ählich
}

Vergessen sie aber nicht, die Data-instanz im Destruktor wieder freizugeben! Der große Vorteil dieser Vor-
gehensweise ist, dass sich das Klasseninterface nicht ändert, wenn z.B. noch weitere Daten-elemente zu der
Data-Klasse hinzugefügt werden. Damit ist das Klasseninterface komplett unabhängig von dem Internen Daten-
Management. 
********************************************************************************************************** */
